﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Threading;

class Program {
	static void RunTasks(bool cancel) {
		CancellationTokenSource cts = new CancellationTokenSource();
		CancellationToken ct = cts.Token;

		var t1 = Task.Run(() => {
			Console.WriteLine("Task 1 (in thread {0}): starting wait via Task.Delay().Wait()", Thread.CurrentThread.ManagedThreadId);

			Task.Delay(2000, ct).Wait();

			Console.WriteLine("Task 1 (in thread {0}): done waiting ...", Thread.CurrentThread.ManagedThreadId);
		}, ct);

		Thread.Sleep(100);

		var t2 = Task.Run(() => {
			Console.WriteLine("Task 2 (in thread {0}): starting wait via Thread.Sleep()", Thread.CurrentThread.ManagedThreadId);

			Thread.Sleep(2000);

			Console.WriteLine("Task 2 (in thread {0}): done waiting ...", Thread.CurrentThread.ManagedThreadId);
		}, ct);

		Thread.Sleep(100);

		var t3 = Task.Run(() => {
			Console.WriteLine("Task 3 (in thread {0}): starting wait via Task.Delay().ContinueWith()", Thread.CurrentThread.ManagedThreadId);

			Task.Delay(3000, ct).ContinueWith(_ => {
				Console.WriteLine("Task 3 (in thread {0}): done waiting ...", Thread.CurrentThread.ManagedThreadId);
			}, ct);
		}, ct);

		var t4 = Task.Run(() => {
			Console.WriteLine("Task 4 (in thread {0}): starting wait via Task.Delay().ContinueWith(AttachedToParent)", Thread.CurrentThread.ManagedThreadId);

			Task.Delay(2000, ct).ContinueWith(_ => {
				Console.WriteLine("Task 4 (in thread {0}): done waiting ...", Thread.CurrentThread.ManagedThreadId);
			}, ct, TaskContinuationOptions.AttachedToParent, TaskScheduler.Current);	// Same as in task 3, as task 4 is not our parent!
		}, ct);

		if (cancel) {
			cts.Cancel();
		}

		try {
			Task.WaitAll(t1, t2, t3, t4);
		} catch (Exception ex) {
			Console.WriteLine("Main: exception caught from WaitAll:");
			Console.WriteLine(ex);
			if (ex is AggregateException) {

				Console.WriteLine("Main: InnerExceptions.Count: {0}", ((AggregateException) ex).InnerExceptions.Count);
			}
		}

		Console.WriteLine("Main: WaitAll completed ... entering sleep ...");
		Thread.Sleep(4000);
		Console.WriteLine("Main: done sleeping, exiting ...");
	}

	static void Main(string[] args) {
		RunTasks(false);
		Console.WriteLine("--------------");
		Console.Write("Press ENTER to repeat test with cancellation ...");
		Console.ReadLine();
		RunTasks(true);
	}
}

